@@ -19,6 +19,13 @@ func Test(t *testing.T) {
 		"fmtappendf",
 		"mapsloop",
 		"minmax",
+		"omitzero",
+		"rangeint",
+		"slicescontains",
+		"slicesdelete",
+		"splitseq",
+		"fieldsseq",
 		"sortslice",
+		"testingcontext",
 	)
 }
