@@ -16,6 +16,8 @@ import (
 	"golang.org/x/tools/gopls/internal/util/moreslices"
 	"golang.org/x/tools/internal/analysisinternal"
 	"golang.org/x/tools/internal/astutil/cursor"
+	"golang.org/x/tools/internal/astutil/edge"
+	"golang.org/x/tools/internal/typesinternal"
 )
 
 //go:embed doc.go
@@ -137,59 +139,72 @@ func run(pass *analysis.Pass) (any, error) {
 		}
 	}
 
-	// Check each non-address-taken function's parameters are all used.
-	filter := []ast.Node{
-		(*ast.FuncDecl)(nil),
-		(*ast.FuncLit)(nil),
-	}
-	cursor.Root(inspect).Inspect(filter, func(c cursor.Cursor, push bool) bool {
-		// (We always return true so that we visit nested FuncLits.)
-
+	// Inspect each file to see if it is generated.
+	//
+	// We do not want to report unused parameters in generated code itself,
+	// however we need to include generated code in the overall analysis as
+	// it may be calling functions in non-generated code.
+	files := []ast.Node{(*ast.File)(nil)}
+	cursor.Root(inspect).Inspect(files, func(c cursor.Cursor, push bool) bool {
 		if !push {
 			return true
 		}
 
-		var (
-			fn    types.Object // function symbol (*Func, possibly *Var for a FuncLit)
-			ftype *ast.FuncType
-			body  *ast.BlockStmt
-		)
-		switch n := c.Node().(type) {
-		case *ast.FuncDecl:
-			// We can't analyze non-Go functions.
-			if n.Body == nil {
+		isGenerated := ast.IsGenerated(c.Node().(*ast.File))
+
+		// Descend into the file, check each non-address-taken function's parameters
+		// are all used.
+		funcs := []ast.Node{
+			(*ast.FuncDecl)(nil),
+			(*ast.FuncLit)(nil),
+		}
+		c.Inspect(funcs, func(c cursor.Cursor, push bool) bool {
+			// (We always return true so that we visit nested FuncLits.)
+			if !push {
 				return true
 			}
 
-			// Ignore exported functions and methods: we
-			// must assume they may be address-taken in
-			// another package.
-			if n.Name.IsExported() {
-				return true
-			}
+			var (
+				fn    types.Object // function symbol (*Func, possibly *Var for a FuncLit)
+				ftype *ast.FuncType
+				body  *ast.BlockStmt
+			)
+			switch n := c.Node().(type) {
+			case *ast.FuncDecl:
+				// We can't analyze non-Go functions.
+				if n.Body == nil {
+					return true
+				}
 
-			// Ignore methods that match the name of any
-			// interface method declared in this package,
-			// as the method's signature may need to conform
-			// to the interface.
-			if n.Recv != nil && unexportedIMethodNames[n.Name.Name] {
-				return true
-			}
+				// Ignore exported functions and methods: we
+				// must assume they may be address-taken in
+				// another package.
+				if n.Name.IsExported() {
+					return true
+				}
 
-			fn = pass.TypesInfo.Defs[n.Name].(*types.Func)
-			ftype, body = n.Type, n.Body
+				// Ignore methods that match the name of any
+				// interface method declared in this package,
+				// as the method's signature may need to conform
+				// to the interface.
+				if n.Recv != nil && unexportedIMethodNames[n.Name.Name] {
+					return true
+				}
 
-		case *ast.FuncLit:
-			// Find the symbol for the variable (if any)
-			// to which the FuncLit is bound.
-			// (We don't bother to allow ParenExprs.)
-			switch parent := c.Parent().Node().(type) {
-			case *ast.AssignStmt:
-				// f  = func() {...}
-				// f := func() {...}
-				for i, rhs := range parent.Rhs {
-					if rhs == n {
-						if id, ok := parent.Lhs[i].(*ast.Ident); ok {
+				fn = pass.TypesInfo.Defs[n.Name].(*types.Func)
+				ftype, body = n.Type, n.Body
+
+			case *ast.FuncLit:
+				// Find the symbol for the variable (if any)
+				// to which the FuncLit is bound.
+				// (We don't bother to allow ParenExprs.)
+				switch parent := c.Parent().Node().(type) {
+				case *ast.AssignStmt:
+					// f  = func() {...}
+					// f := func() {...}
+					if e, idx := c.Edge(); e == edge.AssignStmt_Rhs {
+						// Inv: n == AssignStmt.Rhs[idx]
+						if id, ok := parent.Lhs[idx].(*ast.Ident); ok {
 							fn = pass.TypesInfo.ObjectOf(id)
 
 							// Edge case: f = func() {...}
@@ -200,109 +215,117 @@ func run(pass *analysis.Pass) (any, error) {
 
 							if fn == nil && id.Name == "_" {
 								// Edge case: _ = func() {...}
-								// has no var. Fake one.
-								fn = types.NewVar(id.Pos(), pass.Pkg, id.Name, pass.TypesInfo.TypeOf(n))
+								// has no local var. Fake one.
+								v := types.NewVar(id.Pos(), pass.Pkg, id.Name, pass.TypesInfo.TypeOf(n))
+								typesinternal.SetVarKind(v, typesinternal.LocalVar)
+								fn = v
 							}
 						}
-						break
 					}
-				}
 
-			case *ast.ValueSpec:
-				// var f = func() { ... }
-				// (unless f is an exported package-level var)
-				for i, val := range parent.Values {
-					if val == n {
-						v := pass.TypesInfo.Defs[parent.Names[i]]
-						if !(v.Parent() == pass.Pkg.Scope() && v.Exported()) {
-							fn = v
+				case *ast.ValueSpec:
+					// var f = func() { ... }
+					// (unless f is an exported package-level var)
+					for i, val := range parent.Values {
+						if val == n {
+							v := pass.TypesInfo.Defs[parent.Names[i]]
+							if !(v.Parent() == pass.Pkg.Scope() && v.Exported()) {
+								fn = v
+							}
+							break
+						}
+					}
+				}
+
+				ftype, body = n.Type, n.Body
+			}
+
+			// Ignore address-taken functions and methods: unused
+			// parameters may be needed to conform to a func type.
+			if fn == nil || len(usesOutsideCall[fn]) > 0 {
+				return true
+			}
+
+			// If there are no parameters, there are no unused parameters.
+			if ftype.Params.NumFields() == 0 {
+				return true
+			}
+
+			// To reduce false positives, ignore functions with an
+			// empty or panic body.
+			//
+			// We choose not to ignore functions whose body is a
+			// single return statement (as earlier versions did)
+			// 	func f() { return }
+			// 	func f() { return g(...) }
+			// as we suspect that was just heuristic to reduce
+			// false positives in the earlier unsound algorithm.
+			switch len(body.List) {
+			case 0:
+				// Empty body. Although the parameter is
+				// unnecessary, it's pretty obvious to the
+				// reader that that's the case, so we allow it.
+				return true // func f() {}
+			case 1:
+				if stmt, ok := body.List[0].(*ast.ExprStmt); ok {
+					// We allow a panic body, as it is often a
+					// placeholder for a future implementation:
+					// 	func f() { panic(...) }
+					if call, ok := stmt.X.(*ast.CallExpr); ok {
+						if fun, ok := call.Fun.(*ast.Ident); ok && fun.Name == "panic" {
+							return true
 						}
-						break
 					}
 				}
 			}
 
-			ftype, body = n.Type, n.Body
-		}
-
-		// Ignore address-taken functions and methods: unused
-		// parameters may be needed to conform to a func type.
-		if fn == nil || len(usesOutsideCall[fn]) > 0 {
-			return true
-		}
-
-		// If there are no parameters, there are no unused parameters.
-		if ftype.Params.NumFields() == 0 {
-			return true
-		}
-
-		// To reduce false positives, ignore functions with an
-		// empty or panic body.
-		//
-		// We choose not to ignore functions whose body is a
-		// single return statement (as earlier versions did)
-		// 	func f() { return }
-		// 	func f() { return g(...) }
-		// as we suspect that was just heuristic to reduce
-		// false positives in the earlier unsound algorithm.
-		switch len(body.List) {
-		case 0:
-			// Empty body. Although the parameter is
-			// unnecessary, it's pretty obvious to the
-			// reader that that's the case, so we allow it.
-			return true // func f() {}
-		case 1:
-			if stmt, ok := body.List[0].(*ast.ExprStmt); ok {
-				// We allow a panic body, as it is often a
-				// placeholder for a future implementation:
-				// 	func f() { panic(...) }
-				if call, ok := stmt.X.(*ast.CallExpr); ok {
-					if fun, ok := call.Fun.(*ast.Ident); ok && fun.Name == "panic" {
-						return true
-					}
-				}
+			// Don't report diagnostics on generated files.
+			if isGenerated {
+				return true
 			}
-		}
 
-		// Report each unused parameter.
-		for _, field := range ftype.Params.List {
-			for _, id := range field.Names {
-				if id.Name == "_" {
-					continue
-				}
-				param := pass.TypesInfo.Defs[id].(*types.Var)
-				if !usedVars[param] {
-					start, end := field.Pos(), field.End()
-					if len(field.Names) > 1 {
-						start, end = id.Pos(), id.End()
+			// Report each unused parameter.
+			for _, field := range ftype.Params.List {
+				for _, id := range field.Names {
+					if id.Name == "_" {
+						continue
 					}
-					// This diagnostic carries both an edit-based fix to
-					// rename the unused parameter, and a command-based fix
-					// to remove it (see golang.RemoveUnusedParameter).
-					pass.Report(analysis.Diagnostic{
-						Pos:      start,
-						End:      end,
-						Message:  fmt.Sprintf("unused parameter: %s", id.Name),
-						Category: FixCategory,
-						SuggestedFixes: []analysis.SuggestedFix{
-							{
-								Message: `Rename parameter to "_"`,
-								TextEdits: []analysis.TextEdit{{
-									Pos:     id.Pos(),
-									End:     id.End(),
-									NewText: []byte("_"),
-								}},
+					param := pass.TypesInfo.Defs[id].(*types.Var)
+					if !usedVars[param] {
+						start, end := field.Pos(), field.End()
+						if len(field.Names) > 1 {
+							start, end = id.Pos(), id.End()
+						}
+
+						// This diagnostic carries both an edit-based fix to
+						// rename the unused parameter, and a command-based fix
+						// to remove it (see golang.RemoveUnusedParameter).
+						pass.Report(analysis.Diagnostic{
+							Pos:      start,
+							End:      end,
+							Message:  fmt.Sprintf("unused parameter: %s", id.Name),
+							Category: FixCategory,
+							SuggestedFixes: []analysis.SuggestedFix{
+								{
+									Message: `Rename parameter to "_"`,
+									TextEdits: []analysis.TextEdit{{
+										Pos:     id.Pos(),
+										End:     id.End(),
+										NewText: []byte("_"),
+									}},
+								},
+								{
+									Message: fmt.Sprintf("Remove unused parameter %q", id.Name),
+									// No TextEdits => computed by gopls command
+								},
 							},
-							{
-								Message: fmt.Sprintf("Remove unused parameter %q", id.Name),
-								// No TextEdits => computed by gopls command
-							},
-						},
-					})
+						})
+					}
 				}
 			}
-		}
 
+			return true
+		})
 		return true
 	})
 	return nil, nil
