@@ -108,6 +108,12 @@ var C int
 -- c/c2.go --
 package c
 var C int
+-- d/d.go --
+package d
+
+import "io/ioutil"
+
+var _ = ioutil.ReadFile
 `)
 
 	// no files
@@ -141,6 +147,22 @@ var C int
 		res.checkStdout(`c2.go:2:5-6: C redeclared in this block`)
 		res.checkStdout(`c.go:2:5-6: - other declaration of C`)
 	}
+
+	// No deprecated (hint) diagnostic without -severity.
+	{
+		res := gopls(t, tree, "check", "./d/d.go")
+		res.checkExit(true)
+		if len(res.stdout) > 0 {
+			t.Errorf("check ./d/d.go returned unexpected output:\n%s", res.stdout)
+		}
+	}
+
+	// Deprecated (hint) diagnostics with -severity=hint
+	{
+		res := gopls(t, tree, "check", "-severity=hint", "./d/d.go")
+		res.checkExit(true)
+		res.checkStdout(`ioutil.ReadFile is deprecated`)
+	}
 }
 
 // TestCallHierarchy tests the 'call_hierarchy' subcommand (call_hierarchy.go).
@@ -213,8 +235,8 @@ func TestFail(t *testing.T) { t.Fatal("fail") }
 	{
 		res := gopls(t, tree, "codelens", "./a/a_test.go")
 		res.checkExit(true)
-		res.checkStdout(`a_test.go:3: "run test" \[gopls.test\]`)
-		res.checkStdout(`a_test.go:4: "run test" \[gopls.test\]`)
+		res.checkStdout(`a_test.go:3: "run test" \[gopls.run_tests\]`)
+		res.checkStdout(`a_test.go:4: "run test" \[gopls.run_tests\]`)
 	}
 	// no codelens with title/position
 	{
@@ -930,7 +952,7 @@ package foo
 		res3 := goplsWithEnv(t, tree, []string{"GOPACKAGESDRIVER=off"}, "stats", "-anon")
 		res3.checkExit(true)
 
-		var statsAsMap3 map[string]interface{}
+		var statsAsMap3 map[string]any
 		if err := json.Unmarshal([]byte(res3.stdout), &statsAsMap3); err != nil {
 			t.Fatalf("failed to unmarshal JSON output of stats command: %v", err)
 		}
@@ -950,12 +972,12 @@ func TestCodeAction(t *testing.T) {
 module example.com
 go 1.18
 
--- a.go --
+-- a/a.go --
 package a
 type T int
 func f() (int, string) { return }
 
--- b.go --
+-- a/b.go --
 package a
 import "io"
 var _ io.Reader = C{}
@@ -970,14 +992,14 @@ type C struct{}
 	}
 	// list code actions in file
 	{
-		res := gopls(t, tree, "codeaction", "a.go")
+		res := gopls(t, tree, "codeaction", "a/a.go")
 		res.checkExit(true)
 		res.checkStdout(`edit	"Fill in return values" \[quickfix\]`)
 		res.checkStdout(`command	"Browse documentation for package a" \[source.doc\]`)
 	}
 	// list code actions in file, filtering by title
 	{
-		res := gopls(t, tree, "codeaction", "-title=Browse.*doc", "a.go")
+		res := gopls(t, tree, "codeaction", "-title=Browse.*doc", "a/a.go")
 		res.checkExit(true)
 		got := res.stdout
 		want := `command	"Browse gopls feature documentation" [gopls.doc.features]` +
@@ -990,12 +1012,12 @@ type C struct{}
 	}
 	// list code actions in file, filtering (hierarchically) by kind
 	{
-		res := gopls(t, tree, "codeaction", "-kind=source", "a.go")
+		res := gopls(t, tree, "codeaction", "-kind=source", "a/a.go")
 		res.checkExit(true)
 		got := res.stdout
 		want := `command	"Browse documentation for package a" [source.doc]` +
 			"\n" +
-			`command	"Toggle compiler optimization details" [source.toggleCompilerOptDetails]` +
+			`command	"Show compiler optimization details for \"a\"" [source.toggleCompilerOptDetails]` +
 			"\n"
 		if got != want {
 			t.Errorf("codeaction: got <<%s>>, want <<%s>>\nstderr:\n%s", got, want, res.stderr)
@@ -1003,13 +1025,13 @@ type C struct{}
 	}
 	// list code actions at position (of io.Reader)
 	{
-		res := gopls(t, tree, "codeaction", "b.go:#31")
+		res := gopls(t, tree, "codeaction", "a/b.go:#31")
 		res.checkExit(true)
 		res.checkStdout(`command	"Browse documentation for type io.Reader" \[source.doc]`)
 	}
 	// list quick fixes at position (of type T)
 	{
-		res := gopls(t, tree, "codeaction", "-kind=quickfix", "a.go:#15")
+		res := gopls(t, tree, "codeaction", "-kind=quickfix", "a/a.go:#15")
 		res.checkExit(true)
 		got := res.stdout
 		want := `edit	"Fill in return values" [quickfix]` + "\n"
@@ -1019,7 +1041,7 @@ type C struct{}
 	}
 	// success, with explicit CodeAction kind and diagnostics span.
 	{
-		res := gopls(t, tree, "codeaction", "-kind=quickfix", "-exec", "b.go:#40")
+		res := gopls(t, tree, "codeaction", "-kind=quickfix", "-exec", "a/b.go:#40")
 		res.checkExit(true)
 		got := res.stdout
 		want := `
@@ -1212,7 +1234,7 @@ func (res *result) checkOutput(pattern, name, content string) {
 }
 
 // toJSON decodes res.stdout as JSON into to *ptr and reports its success.
-func (res *result) toJSON(ptr interface{}) bool {
+func (res *result) toJSON(ptr any) bool {
 	if err := json.Unmarshal([]byte(res.stdout), ptr); err != nil {
 		res.t.Errorf("invalid JSON %v", err)
 		return false
