@@ -11,6 +11,7 @@ import (
 	"go/ast"
 	"go/token"
 	"go/types"
+	"path/filepath"
 	"reflect"
 	"slices"
 	"sort"
@@ -27,7 +28,6 @@ import (
 	"golang.org/x/tools/gopls/internal/protocol"
 	"golang.org/x/tools/gopls/internal/protocol/command"
 	"golang.org/x/tools/gopls/internal/settings"
-	"golang.org/x/tools/gopls/internal/util/typesutil"
 	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/imports"
 	"golang.org/x/tools/internal/typesinternal"
@@ -106,6 +106,8 @@ func CodeActions(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle,
 			req.pkg = nil
 		}
 		if err := p.fn(ctx, req); err != nil {
+			// TODO(adonovan): most errors in code action providers should
+			// not block other providers; see https://go.dev/issue/71275.
 			return nil, err
 		}
 	}
@@ -152,7 +154,14 @@ func (req *codeActionsRequest) addApplyFixAction(title, fix string, loc protocol
 // then the command is embedded into the code action data field so
 // that the client can later ask the server to "resolve" a command
 // into an edit that they can preview and apply selectively.
-// Set allowResolveEdits only for actions that generate edits.
+// IMPORTANT: set allowResolveEdits only for actions that are 'edit aware',
+// meaning they can detect when they are being executed in the context of a
+// codeAction/resolve request, and return edits rather than applying them using
+// workspace/applyEdit. In golang/go#71405, edits were being apply during the
+// codeAction/resolve request handler.
+// TODO(rfindley): refactor the command and code lens registration APIs so that
+// resolve edit support is inferred from the command signature, not dependent
+// on coordination between codeAction and command logic.
 //
 // Otherwise, the command is set as the code action operation.
 func (req *codeActionsRequest) addCommandAction(cmd *protocol.Command, allowResolveEdits bool) {
@@ -173,7 +182,7 @@ func (req *codeActionsRequest) addCommandAction(cmd *protocol.Command, allowReso
 	req.addAction(act)
 }
 
-// addCommandAction adds an edit-based CodeAction to the result.
+// addEditAction adds an edit-based CodeAction to the result.
 func (req *codeActionsRequest) addEditAction(title string, fixedDiagnostics []protocol.Diagnostic, changes ...protocol.DocumentChange) {
 	req.addAction(protocol.CodeAction{
 		Title:       title,
@@ -227,15 +236,20 @@ type codeActionProducer struct {
 var codeActionProducers = [...]codeActionProducer{
 	{kind: protocol.QuickFix, fn: quickFix, needPkg: true},
 	{kind: protocol.SourceOrganizeImports, fn: sourceOrganizeImports},
+	{kind: settings.AddTest, fn: addTest, needPkg: true},
 	{kind: settings.GoAssembly, fn: goAssembly, needPkg: true},
 	{kind: settings.GoDoc, fn: goDoc, needPkg: true},
 	{kind: settings.GoFreeSymbols, fn: goFreeSymbols},
 	{kind: settings.GoTest, fn: goTest},
+	{kind: settings.GoToggleCompilerOptDetails, fn: toggleCompilerOptDetails},
 	{kind: settings.GoplsDocFeatures, fn: goplsDocFeatures},
 	{kind: settings.RefactorExtractFunction, fn: refactorExtractFunction},
 	{kind: settings.RefactorExtractMethod, fn: refactorExtractMethod},
 	{kind: settings.RefactorExtractToNewFile, fn: refactorExtractToNewFile},
-	{kind: settings.RefactorExtractVariable, fn: refactorExtractVariable},
+	{kind: settings.RefactorExtractConstant, fn: refactorExtractVariable, needPkg: true},
+	{kind: settings.RefactorExtractVariable, fn: refactorExtractVariable, needPkg: true},
+	{kind: settings.RefactorExtractConstantAll, fn: refactorExtractVariableAll, needPkg: true},
+	{kind: settings.RefactorExtractVariableAll, fn: refactorExtractVariableAll, needPkg: true},
 	{kind: settings.RefactorInlineCall, fn: refactorInlineCall, needPkg: true},
 	{kind: settings.RefactorRewriteChangeQuote, fn: refactorRewriteChangeQuote},
 	{kind: settings.RefactorRewriteFillStruct, fn: refactorRewriteFillStruct, needPkg: true},
@@ -243,7 +257,10 @@ var codeActionProducers = [...]codeActionProducer{
 	{kind: settings.RefactorRewriteInvertIf, fn: refactorRewriteInvertIf},
 	{kind: settings.RefactorRewriteJoinLines, fn: refactorRewriteJoinLines, needPkg: true},
 	{kind: settings.RefactorRewriteRemoveUnusedParam, fn: refactorRewriteRemoveUnusedParam, needPkg: true},
+	{kind: settings.RefactorRewriteMoveParamLeft, fn: refactorRewriteMoveParamLeft, needPkg: true},
+	{kind: settings.RefactorRewriteMoveParamRight, fn: refactorRewriteMoveParamRight, needPkg: true},
 	{kind: settings.RefactorRewriteSplitLines, fn: refactorRewriteSplitLines, needPkg: true},
+	{kind: settings.RefactorRewriteEliminateDotImport, fn: refactorRewriteEliminateDotImport, needPkg: true},
 
 	// Note: don't forget to update the allow-list in Server.CodeAction
 	// when adding new query operations like GoTest and GoDoc that
@@ -293,7 +310,7 @@ func quickFix(ctx context.Context, req *codeActionsRequest) error {
 	for _, typeError := range req.pkg.TypeErrors() {
 		// Does type error overlap with CodeAction range?
 		start, end := typeError.Pos, typeError.Pos
-		if _, _, endPos, ok := typesinternal.ReadGo116ErrorData(typeError); ok {
+		if _, _, endPos, ok := typesinternal.ErrorCodeStartEnd(typeError); ok {
 			end = endPos
 		}
 		typeErrorRange, err := req.pgf.PosRange(start, end)
@@ -301,20 +318,41 @@ func quickFix(ctx context.Context, req *codeActionsRequest) error {
 			continue
 		}
 
+		msg := typeError.Msg
+		switch {
 		// "Missing method" error? (stubmethods)
 		// Offer a "Declare missing methods of INTERFACE" code action.
-		// See [stubMethodsFixer] for command implementation.
-		msg := typeError.Error()
-		if strings.Contains(msg, "missing method") ||
-			strings.HasPrefix(msg, "cannot convert") ||
-			strings.Contains(msg, "not implement") {
-			path, _ := astutil.PathEnclosingInterval(req.pgf.File, start, end)
-			si := stubmethods.GetStubInfo(req.pkg.FileSet(), info, path, start)
+		// See [stubMissingInterfaceMethodsFixer] for command implementation.
+		case strings.Contains(msg, "missing method"),
+			strings.HasPrefix(msg, "cannot convert"),
+			strings.Contains(msg, "not implement"):
+			si := stubmethods.GetIfaceStubInfo(req.pkg.FileSet(), info, req.pgf, start, end)
 			if si != nil {
-				qf := typesutil.FileQualifier(req.pgf.File, si.Concrete.Obj().Pkg(), info)
-				iface := types.TypeString(si.Interface.Type(), qf)
+				qual := typesinternal.FileQualifier(req.pgf.File, si.Concrete.Obj().Pkg())
+				iface := types.TypeString(si.Interface.Type(), qual)
 				msg := fmt.Sprintf("Declare missing methods of %s", iface)
-				req.addApplyFixAction(msg, fixStubMethods, req.loc)
+				req.addApplyFixAction(msg, fixMissingInterfaceMethods, req.loc)
+			}
+
+		// "type X has no field or method Y" compiler error.
+		// Offer a "Declare missing method T.f" code action.
+		// See [stubMissingCalledFunctionFixer] for command implementation.
+		case strings.Contains(msg, "has no field or method"):
+			si := stubmethods.GetCallStubInfo(req.pkg.FileSet(), info, req.pgf, start, end)
+			if si != nil {
+				msg := fmt.Sprintf("Declare missing method %s.%s", si.Receiver.Obj().Name(), si.MethodName)
+				req.addApplyFixAction(msg, fixMissingCalledFunction, req.loc)
+			}
+
+		// "undeclared name: X" or "undefined: X" compiler error.
+		// Offer a "Create variable/function X" code action.
+		// See [createUndeclared] for command implementation.
+		case strings.HasPrefix(msg, "undeclared name: "),
+			strings.HasPrefix(msg, "undefined: "):
+			path, _ := astutil.PathEnclosingInterval(req.pgf.File, start, end)
+			title := undeclaredFixTitle(path, msg)
+			if title != "" {
+				req.addApplyFixAction(title, fixCreateUndeclared, req.loc)
 			}
 		}
 	}
@@ -413,15 +451,17 @@ func goplsDocFeatures(ctx context.Context, req *codeActionsRequest) error {
 // See [server.commandHandler.Doc] for command implementation.
 func goDoc(ctx context.Context, req *codeActionsRequest) error {
 	_, _, title := DocFragment(req.pkg, req.pgf, req.start, req.end)
-	cmd := command.NewDocCommand(title, command.DocArgs{Location: req.loc, ShowDocument: true})
-	req.addCommandAction(cmd, false)
+	if title != "" {
+		cmd := command.NewDocCommand(title, command.DocArgs{Location: req.loc, ShowDocument: true})
+		req.addCommandAction(cmd, false)
+	}
 	return nil
 }
 
 // refactorExtractFunction produces "Extract function" code actions.
 // See [extractFunction] for command implementation.
 func refactorExtractFunction(ctx context.Context, req *codeActionsRequest) error {
-	if _, ok, _, _ := canExtractFunction(req.pgf.Tok, req.start, req.end, req.pgf.Src, req.pgf.File); ok {
+	if _, ok, _, _ := canExtractFunction(req.pgf.Tok, req.start, req.end, req.pgf.Src, req.pgf.Cursor); ok {
 		req.addApplyFixAction("Extract function", fixExtractFunction, req.loc)
 	}
 	return nil
@@ -430,17 +470,61 @@ func refactorExtractFunction(ctx context.Context, req *codeActionsRequest) error
 // refactorExtractMethod produces "Extract method" code actions.
 // See [extractMethod] for command implementation.
 func refactorExtractMethod(ctx context.Context, req *codeActionsRequest) error {
-	if _, ok, methodOK, _ := canExtractFunction(req.pgf.Tok, req.start, req.end, req.pgf.Src, req.pgf.File); ok && methodOK {
+	if _, ok, methodOK, _ := canExtractFunction(req.pgf.Tok, req.start, req.end, req.pgf.Src, req.pgf.Cursor); ok && methodOK {
 		req.addApplyFixAction("Extract method", fixExtractMethod, req.loc)
 	}
 	return nil
 }
 
-// refactorExtractVariable produces "Extract variable" code actions.
+// refactorExtractVariable produces "Extract variable|constant" code actions.
 // See [extractVariable] for command implementation.
 func refactorExtractVariable(ctx context.Context, req *codeActionsRequest) error {
-	if _, _, ok, _ := canExtractVariable(req.start, req.end, req.pgf.File); ok {
-		req.addApplyFixAction("Extract variable", fixExtractVariable, req.loc)
+	info := req.pkg.TypesInfo()
+	if exprs, err := canExtractVariable(info, req.pgf.Cursor, req.start, req.end, false); err == nil {
+		// Offer one of refactor.extract.{constant,variable}
+		// based on the constness of the expression; this is a
+		// limitation of the codeActionProducers mechanism.
+		// Beware that future evolutions of the refactorings
+		// may make them diverge to become non-complementary,
+		// for example because "if const x = ...; y {" is illegal.
+		// Same as [refactorExtractVariableAll].
+		constant := info.Types[exprs[0]].Value != nil
+		if (req.kind == settings.RefactorExtractConstant) == constant {
+			title := "Extract variable"
+			if constant {
+				title = "Extract constant"
+			}
+			req.addApplyFixAction(title, fixExtractVariable, req.loc)
+		}
+	}
+	return nil
+}
+
+// refactorExtractVariableAll produces "Extract N occurrences of EXPR" code action.
+// See [extractAllOccursOfExpr] for command implementation.
+func refactorExtractVariableAll(ctx context.Context, req *codeActionsRequest) error {
+	info := req.pkg.TypesInfo()
+	// Don't suggest if only one expr is found,
+	// otherwise it will duplicate with [refactorExtractVariable]
+	if exprs, err := canExtractVariable(info, req.pgf.Cursor, req.start, req.end, true); err == nil && len(exprs) > 1 {
+		start, end, err := req.pgf.NodeOffsets(exprs[0])
+		if err != nil {
+			return err
+		}
+		desc := string(req.pgf.Src[start:end])
+		if len(desc) >= 40 || strings.Contains(desc, "\n") {
+			desc = astutil.NodeDescription(exprs[0])
+		}
+		constant := info.Types[exprs[0]].Value != nil
+		if (req.kind == settings.RefactorExtractConstantAll) == constant {
+			var title string
+			if constant {
+				title = fmt.Sprintf("Extract %d occurrences of const expression: %s", len(exprs), desc)
+			} else {
+				title = fmt.Sprintf("Extract %d occurrences of %s", len(exprs), desc)
+			}
+			req.addApplyFixAction(title, fixExtractVariableAll, req.loc)
+		}
 	}
 	return nil
 }
@@ -450,34 +534,136 @@ func refactorExtractVariable(ctx context.Context, req *codeActionsRequest) error
 func refactorExtractToNewFile(ctx context.Context, req *codeActionsRequest) error {
 	if canExtractToNewFile(req.pgf, req.start, req.end) {
 		cmd := command.NewExtractToNewFileCommand("Extract declarations to new file", req.loc)
-		req.addCommandAction(cmd, true)
+		req.addCommandAction(cmd, false)
 	}
 	return nil
 }
 
+// addTest produces "Add test for FUNC" code actions.
+// See [server.commandHandler.AddTest] for command implementation.
+func addTest(ctx context.Context, req *codeActionsRequest) error {
+	// Reject test package.
+	if req.pkg.Metadata().ForTest != "" {
+		return nil
+	}
+
+	path, _ := astutil.PathEnclosingInterval(req.pgf.File, req.start, req.end)
+	if len(path) < 2 {
+		return nil
+	}
+
+	decl, ok := path[len(path)-2].(*ast.FuncDecl)
+	if !ok {
+		return nil
+	}
+
+	// Don't offer to create tests of "init" or "_".
+	if decl.Name.Name == "_" || decl.Name.Name == "init" {
+		return nil
+	}
+
+	// TODO(hxjiang): support functions with type parameter.
+	if decl.Type.TypeParams != nil {
+		return nil
+	}
+
+	cmd := command.NewAddTestCommand("Add test for "+decl.Name.String(), req.loc)
+	req.addCommandAction(cmd, false)
+
+	// TODO(hxjiang): add code action for generate test for package/file.
+	return nil
+}
+
+// identityTransform returns a change signature transformation that leaves the
+// given fieldlist unmodified.
+func identityTransform(fields *ast.FieldList) []command.ChangeSignatureParam {
+	var id []command.ChangeSignatureParam
+	for i := 0; i < fields.NumFields(); i++ {
+		id = append(id, command.ChangeSignatureParam{OldIndex: i})
+	}
+	return id
+}
+
 // refactorRewriteRemoveUnusedParam produces "Remove unused parameter" code actions.
 // See [server.commandHandler.ChangeSignature] for command implementation.
 func refactorRewriteRemoveUnusedParam(ctx context.Context, req *codeActionsRequest) error {
-	if canRemoveParameter(req.pkg, req.pgf, req.loc.Range) {
-		cmd := command.NewChangeSignatureCommand("Refactor: remove unused parameter", command.ChangeSignatureArgs{
-			RemoveParameter: req.loc,
-			ResolveEdits:    req.resolveEdits(),
+	if info := removableParameter(req.pkg, req.pgf, req.loc.Range); info != nil {
+		var transform []command.ChangeSignatureParam
+		for i := 0; i < info.decl.Type.Params.NumFields(); i++ {
+			if i != info.paramIndex {
+				transform = append(transform, command.ChangeSignatureParam{OldIndex: i})
+			}
+		}
+		cmd := command.NewChangeSignatureCommand("Remove unused parameter", command.ChangeSignatureArgs{
+			Location:     req.loc,
+			NewParams:    transform,
+			NewResults:   identityTransform(info.decl.Type.Results),
+			ResolveEdits: req.resolveEdits(),
 		})
 		req.addCommandAction(cmd, true)
 	}
 	return nil
 }
 
+func refactorRewriteMoveParamLeft(ctx context.Context, req *codeActionsRequest) error {
+	if info := findParam(req.pgf, req.loc.Range); info != nil &&
+		info.paramIndex > 0 &&
+		!is[*ast.Ellipsis](info.field.Type) {
+
+		// ^^ we can't currently handle moving a variadic param.
+		// TODO(rfindley): implement.
+
+		transform := identityTransform(info.decl.Type.Params)
+		transform[info.paramIndex] = command.ChangeSignatureParam{OldIndex: info.paramIndex - 1}
+		transform[info.paramIndex-1] = command.ChangeSignatureParam{OldIndex: info.paramIndex}
+		cmd := command.NewChangeSignatureCommand("Move parameter left", command.ChangeSignatureArgs{
+			Location:     req.loc,
+			NewParams:    transform,
+			NewResults:   identityTransform(info.decl.Type.Results),
+			ResolveEdits: req.resolveEdits(),
+		})
+
+		req.addCommandAction(cmd, true)
+	}
+	return nil
+}
+
+func refactorRewriteMoveParamRight(ctx context.Context, req *codeActionsRequest) error {
+	if info := findParam(req.pgf, req.loc.Range); info != nil && info.paramIndex >= 0 {
+		params := info.decl.Type.Params
+		nparams := params.NumFields()
+		if info.paramIndex < nparams-1 { // not the last param
+			if info.paramIndex == nparams-2 && is[*ast.Ellipsis](params.List[len(params.List)-1].Type) {
+				// We can't currently handle moving a variadic param.
+				// TODO(rfindley): implement.
+				return nil
+			}
+
+			transform := identityTransform(info.decl.Type.Params)
+			transform[info.paramIndex] = command.ChangeSignatureParam{OldIndex: info.paramIndex + 1}
+			transform[info.paramIndex+1] = command.ChangeSignatureParam{OldIndex: info.paramIndex}
+			cmd := command.NewChangeSignatureCommand("Move parameter right", command.ChangeSignatureArgs{
+				Location:     req.loc,
+				NewParams:    transform,
+				NewResults:   identityTransform(info.decl.Type.Results),
+				ResolveEdits: req.resolveEdits(),
+			})
+			req.addCommandAction(cmd, true)
+		}
+	}
+	return nil
+}
+
 // refactorRewriteChangeQuote produces "Convert to {raw,interpreted} string literal" code actions.
 func refactorRewriteChangeQuote(ctx context.Context, req *codeActionsRequest) error {
 	convertStringLiteral(req)
 	return nil
 }
 
-// refactorRewriteChangeQuote produces "Invert 'if' condition" code actions.
+// refactorRewriteInvertIf produces "Invert 'if' condition" code actions.
 // See [invertIfCondition] for command implementation.
 func refactorRewriteInvertIf(ctx context.Context, req *codeActionsRequest) error {
-	if _, ok, _ := canInvertIfCondition(req.pgf.File, req.start, req.end); ok {
+	if _, ok, _ := canInvertIfCondition(req.pgf.Cursor, req.start, req.end); ok {
 		req.addApplyFixAction("Invert 'if' condition", fixInvertIfCondition, req.loc)
 	}
 	return nil
@@ -487,17 +673,116 @@ func refactorRewriteInvertIf(ctx context.Context, req *codeActionsRequest) error
 // See [splitLines] for command implementation.
 func refactorRewriteSplitLines(ctx context.Context, req *codeActionsRequest) error {
 	// TODO(adonovan): opt: don't set needPkg just for FileSet.
-	if msg, ok, _ := canSplitLines(req.pgf.File, req.pkg.FileSet(), req.start, req.end); ok {
+	if msg, ok, _ := canSplitLines(req.pgf.Cursor, req.pkg.FileSet(), req.start, req.end); ok {
 		req.addApplyFixAction(msg, fixSplitLines, req.loc)
 	}
 	return nil
 }
 
+func refactorRewriteEliminateDotImport(ctx context.Context, req *codeActionsRequest) error {
+	// Figure out if the request is placed over a dot import.
+	var importSpec *ast.ImportSpec
+	for _, imp := range req.pgf.File.Imports {
+		if posRangeContains(imp.Pos(), imp.End(), req.start, req.end) {
+			importSpec = imp
+			break
+		}
+	}
+	if importSpec == nil {
+		return nil
+	}
+	if importSpec.Name == nil || importSpec.Name.Name != "." {
+		return nil
+	}
+
+	// dotImported package path and its imported name after removing the dot.
+	imported := req.pkg.TypesInfo().PkgNameOf(importSpec).Imported()
+	newName := imported.Name()
+
+	rng, err := req.pgf.PosRange(importSpec.Name.Pos(), importSpec.Path.Pos())
+	if err != nil {
+		return err
+	}
+	// Delete the '.' part of the import.
+	edits := []protocol.TextEdit{{
+		Range: rng,
+	}}
+
+	fileScope, ok := req.pkg.TypesInfo().Scopes[req.pgf.File]
+	if !ok {
+		return nil
+	}
+
+	// Go through each use of the dot imported package, checking its scope for
+	// shadowing and calculating an edit to qualify the identifier.
+	var stack []ast.Node
+	ast.Inspect(req.pgf.File, func(n ast.Node) bool {
+		if n == nil {
+			stack = stack[:len(stack)-1] // pop
+			return false
+		}
+		stack = append(stack, n) // push
+
+		ident, ok := n.(*ast.Ident)
+		if !ok {
+			return true
+		}
+		// Only keep identifiers that use a symbol from the
+		// dot imported package.
+		use := req.pkg.TypesInfo().Uses[ident]
+		if use == nil || use.Pkg() == nil {
+			return true
+		}
+		if use.Pkg() != imported {
+			return true
+		}
+
+		// Only qualify unqualified identifiers (due to dot imports).
+		// All other references to a symbol imported from another package
+		// are nested within a select expression (pkg.Foo, v.Method, v.Field).
+		if is[*ast.SelectorExpr](stack[len(stack)-2]) {
+			return true
+		}
+
+		// Make sure that the package name will not be shadowed by something else in scope.
+		// If it is then we cannot offer this particular code action.
+		//
+		// TODO: If the object found in scope is the package imported without a
+		// dot, or some builtin not used in the file, the code action could be
+		// allowed to go through.
+		sc := fileScope.Innermost(ident.Pos())
+		if sc == nil {
+			return true
+		}
+		_, obj := sc.LookupParent(newName, ident.Pos())
+		if obj != nil {
+			return true
+		}
+
+		rng, err := req.pgf.PosRange(ident.Pos(), ident.Pos()) // sic, zero-width range before ident
+		if err != nil {
+			return true
+		}
+		edits = append(edits, protocol.TextEdit{
+			Range:   rng,
+			NewText: newName + ".",
+		})
+
+		return true
+	})
+
+	req.addEditAction("Eliminate dot import", nil, protocol.DocumentChangeEdit(
+		req.fh,
+		edits,
+	))
+	return nil
+}
+
 // refactorRewriteJoinLines produces "Join ITEMS into one line" code actions.
 // See [joinLines] for command implementation.
 func refactorRewriteJoinLines(ctx context.Context, req *codeActionsRequest) error {
 	// TODO(adonovan): opt: don't set needPkg just for FileSet.
-	if msg, ok, _ := canJoinLines(req.pgf.File, req.pkg.FileSet(), req.start, req.end); ok {
+	if msg, ok, _ := canJoinLines(req.pgf.Cursor, req.pkg.FileSet(), req.start, req.end); ok {
 		req.addApplyFixAction(msg, fixJoinLines, req.loc)
 	}
 	return nil
@@ -534,10 +819,10 @@ func refactorRewriteFillSwitch(ctx context.Context, req *codeActionsRequest) err
 	return nil
 }
 
-// canRemoveParameter reports whether we can remove the function parameter
-// indicated by the given [start, end) range.
+// removableParameter returns paramInfo about a removable parameter indicated
+// by the given [start, end) range, or nil if no such removal is available.
 //
-// This is true if:
+// Removing a parameter is possible if
 //   - there are no parse or type errors, and
 //   - [start, end) is contained within an unused field or parameter name
 //   - ... of a non-method function declaration.
@@ -546,33 +831,30 @@ func refactorRewriteFillSwitch(ctx context.Context, req *codeActionsRequest) err
 // much more precisely, allowing it to report its findings as diagnostics.)
 //
 // TODO(adonovan): inline into refactorRewriteRemoveUnusedParam.
-func canRemoveParameter(pkg *cache.Package, pgf *parsego.File, rng protocol.Range) bool {
+func removableParameter(pkg *cache.Package, pgf *parsego.File, rng protocol.Range) *paramInfo {
 	if perrors, terrors := pkg.ParseErrors(), pkg.TypeErrors(); len(perrors) > 0 || len(terrors) > 0 {
-		return false // can't remove parameters from packages with errors
+		return nil // can't remove parameters from packages with errors
 	}
-	info, err := findParam(pgf, rng)
-	if err != nil {
-		return false // e.g. invalid range
-	}
-	if info.field == nil {
-		return false // range does not span a parameter
+	info := findParam(pgf, rng)
+	if info == nil || info.field == nil {
+		return nil // range does not span a parameter
 	}
 	if info.decl.Body == nil {
-		return false // external function
+		return nil // external function
 	}
 	if len(info.field.Names) == 0 {
-		return true // no names => field is unused
+		return info // no names => field is unused
 	}
 	if info.name == nil {
-		return false // no name is indicated
+		return nil // no name is indicated
 	}
 	if info.name.Name == "_" {
-		return true // trivially unused
+		return info // trivially unused
 	}
 
 	obj := pkg.TypesInfo().Defs[info.name]
 	if obj == nil {
-		return false // something went wrong
+		return nil // something went wrong
 	}
 
 	used := false
@@ -582,7 +864,10 @@ func canRemoveParameter(pkg *cache.Package, pgf *parsego.File, rng protocol.Rang
 		}
 		return !used // keep going until we find a use
 	})
-	return !used
+	if used {
+		return nil
+	}
+	return info
 }
 
 // refactorInlineCall produces "Inline call to FUNC" code actions.
@@ -626,7 +911,11 @@ func goTest(ctx context.Context, req *codeActionsRequest) error {
 		return nil
 	}
 
-	cmd := command.NewTestCommand("Run tests and benchmarks", req.loc.URI, tests, benchmarks)
+	cmd := command.NewRunTestsCommand("Run tests and benchmarks", command.RunTestsArgs{
+		URI:        req.loc.URI,
+		Tests:      tests,
+		Benchmarks: benchmarks,
+	})
 	req.addCommandAction(cmd, false)
 	return nil
 }
@@ -656,44 +945,86 @@ func goAssembly(ctx context.Context, req *codeActionsRequest) error {
 	// directly to (say) a lambda of interest.
 	// Perhaps we could scroll to STEXT for the innermost
 	// enclosing nested function?
-	path, _ := astutil.PathEnclosingInterval(req.pgf.File, req.start, req.end)
-	if len(path) >= 2 { // [... FuncDecl File]
-		if decl, ok := path[len(path)-2].(*ast.FuncDecl); ok {
-			if fn, ok := req.pkg.TypesInfo().Defs[decl.Name].(*types.Func); ok {
-				sig := fn.Signature()
 
-				// Compute the linker symbol of the enclosing function.
-				var sym strings.Builder
-				if fn.Pkg().Name() == "main" {
-					sym.WriteString("main")
-				} else {
-					sym.WriteString(fn.Pkg().Path())
-				}
-				sym.WriteString(".")
-				if sig.Recv() != nil {
-					if isPtr, named := typesinternal.ReceiverNamed(sig.Recv()); named != nil {
-						sym.WriteString("(")
-						if isPtr {
-							sym.WriteString("*")
+	// Compute the linker symbol of the enclosing function or var initializer.
+	var sym strings.Builder
+	if pkg := req.pkg.Types(); pkg.Name() == "main" {
+		sym.WriteString("main")
+	} else {
+		sym.WriteString(pkg.Path())
+	}
+	sym.WriteString(".")
+
+	curSel, _ := req.pgf.Cursor.FindPos(req.start, req.end)
+	for cur := range curSel.Ancestors((*ast.FuncDecl)(nil), (*ast.ValueSpec)(nil)) {
+		var name string // in command title
+		switch node := cur.Node().(type) {
+		case *ast.FuncDecl:
+			// package-level func or method
+			if fn, ok := req.pkg.TypesInfo().Defs[node.Name].(*types.Func); ok &&
+				fn.Name() != "_" { // blank functions are not compiled
+
+				// Source-level init functions are compiled (along with
+				// package-level var initializers) in into a single pkg.init
+				// function, so this falls out of the logic below.
+
+				if sig := fn.Signature(); sig.TypeParams() == nil && sig.RecvTypeParams() == nil { // generic => no assembly
+					if sig.Recv() != nil {
+						if isPtr, named := typesinternal.ReceiverNamed(sig.Recv()); named != nil {
+							if isPtr {
+								fmt.Fprintf(&sym, "(*%s)", named.Obj().Name())
+							} else {
+								sym.WriteString(named.Obj().Name())
+							}
+							sym.WriteByte('.')
 						}
-						sym.WriteString(named.Obj().Name())
-						sym.WriteString(").")
 					}
-				}
-				sym.WriteString(fn.Name())
+					sym.WriteString(fn.Name())
 
-				if fn.Name() != "_" && // blank functions are not compiled
-					(fn.Name() != "init" || sig.Recv() != nil) && // init functions aren't linker functions
-					sig.TypeParams() == nil && sig.RecvTypeParams() == nil { // generic => no assembly
-					cmd := command.NewAssemblyCommand(
-						fmt.Sprintf("Browse %s assembly for %s", view.GOARCH(), decl.Name),
-						view.ID(),
-						string(req.pkg.Metadata().ID),
-						sym.String())
-					req.addCommandAction(cmd, false)
+					name = node.Name.Name // success
 				}
 			}
+
+		case *ast.ValueSpec:
+			// package-level var initializer?
+			if len(node.Names) > 0 && len(node.Values) > 0 {
+				v := req.pkg.TypesInfo().Defs[node.Names[0]]
+				if v != nil && typesinternal.IsPackageLevel(v) {
+					sym.WriteString("init")
+					name = "package initializer" // success
+				}
+			}
+		}
+
+		if name != "" {
+			cmd := command.NewAssemblyCommand(
+				fmt.Sprintf("Browse %s assembly for %s", view.GOARCH(), name),
+				view.ID(),
+				string(req.pkg.Metadata().ID),
+				sym.String())
+			req.addCommandAction(cmd, false)
+			break
 		}
 	}
 	return nil
 }
+
+// toggleCompilerOptDetails produces "{Show,Hide} compiler optimization details" code action.
+// See [server.commandHandler.GCDetails] for command implementation.
+func toggleCompilerOptDetails(ctx context.Context, req *codeActionsRequest) error {
+	// TODO(adonovan): errors from code action providers should probably be
+	// logged, even if they aren't visible to the client; see https://go.dev/issue/71275.
+	if meta, err := NarrowestMetadataForFile(ctx, req.snapshot, req.fh.URI()); err == nil {
+		if len(meta.CompiledGoFiles) == 0 {
+			return fmt.Errorf("package %q does not compile file %q", meta.ID, req.fh.URI())
+		}
+		dir := meta.CompiledGoFiles[0].Dir()
+
+		title := fmt.Sprintf("%s compiler optimization details for %q",
+			cond(req.snapshot.WantCompilerOptDetails(dir), "Hide", "Show"),
+			filepath.Base(dir.Path()))
+		cmd := command.NewGCDetailsCommand(title, req.fh.URI())
+		req.addCommandAction(cmd, false)
+	}
+	return nil
+}
