@@ -9,12 +9,13 @@ import (
 	"fmt"
 
 	"golang.org/x/tools/gopls/internal/file"
-	"golang.org/x/tools/gopls/internal/lsp/protocol"
-	"golang.org/x/tools/gopls/internal/lsp/source"
+	"golang.org/x/tools/gopls/internal/goasm"
+	"golang.org/x/tools/gopls/internal/golang"
+	"golang.org/x/tools/gopls/internal/label"
+	"golang.org/x/tools/gopls/internal/protocol"
 	"golang.org/x/tools/gopls/internal/telemetry"
 	"golang.org/x/tools/gopls/internal/template"
 	"golang.org/x/tools/internal/event"
-	"golang.org/x/tools/internal/event/tag"
 )
 
 func (s *server) Definition(ctx context.Context, params *protocol.DefinitionParams) (_ []protocol.Location, rerr error) {
@@ -23,7 +24,7 @@ func (s *server) Definition(ctx context.Context, params *protocol.DefinitionPara
 		recordLatency(ctx, rerr)
 	}()
 
-	ctx, done := event.Start(ctx, "lsp.Server.definition", tag.URI.Of(params.TextDocument.URI))
+	ctx, done := event.Start(ctx, "lsp.Server.definition", label.URI.Of(params.TextDocument.URI))
 	defer done()
 
 	// TODO(rfindley): definition requests should be multiplexed across all views.
@@ -36,14 +37,16 @@ func (s *server) Definition(ctx context.Context, params *protocol.DefinitionPara
 	case file.Tmpl:
 		return template.Definition(snapshot, fh, params.Position)
 	case file.Go:
-		return source.Definition(ctx, snapshot, fh, params.Position)
+		return golang.Definition(ctx, snapshot, fh, params.Position)
+	case file.Asm:
+		return goasm.Definition(ctx, snapshot, fh, params.Position)
 	default:
 		return nil, fmt.Errorf("can't find definitions for file type %s", kind)
 	}
 }
 
 func (s *server) TypeDefinition(ctx context.Context, params *protocol.TypeDefinitionParams) ([]protocol.Location, error) {
-	ctx, done := event.Start(ctx, "lsp.Server.typeDefinition", tag.URI.Of(params.TextDocument.URI))
+	ctx, done := event.Start(ctx, "lsp.Server.typeDefinition", label.URI.Of(params.TextDocument.URI))
 	defer done()
 
 	// TODO(rfindley): type definition requests should be multiplexed across all views.
@@ -54,7 +57,7 @@ func (s *server) TypeDefinition(ctx context.Context, params *protocol.TypeDefini
 	defer release()
 	switch kind := snapshot.FileKind(fh); kind {
 	case file.Go:
-		return source.TypeDefinition(ctx, snapshot, fh, params.Position)
+		return golang.TypeDefinition(ctx, snapshot, fh, params.Position)
 	default:
 		return nil, fmt.Errorf("can't find type definitions for file type %s", kind)
 	}
