@@ -26,10 +26,13 @@ import (
 	debuglog "golang.org/x/tools/gopls/internal/debug/log"
 	"golang.org/x/tools/gopls/internal/file"
 	"golang.org/x/tools/gopls/internal/protocol"
+	"golang.org/x/tools/gopls/internal/protocol/semtok"
 	"golang.org/x/tools/gopls/internal/settings"
+	"golang.org/x/tools/gopls/internal/telemetry"
 	"golang.org/x/tools/gopls/internal/util/bug"
 	"golang.org/x/tools/gopls/internal/util/goversion"
 	"golang.org/x/tools/gopls/internal/util/moremaps"
+	"golang.org/x/tools/gopls/internal/util/moreslices"
 	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/jsonrpc2"
 )
@@ -72,7 +75,11 @@ func (s *server) Initialize(ctx context.Context, params *protocol.ParamInitializ
 	// TODO(rfindley): eliminate this defer.
 	defer func() { s.SetOptions(options) }()
 
-	s.handleOptionErrors(ctx, options.Set(params.InitializationOptions))
+	// Process initialization options.
+	{
+		res, errs := options.Set(params.InitializationOptions)
+		s.handleOptionResult(ctx, res, errs)
+	}
 	options.ForClientCapabilities(params.ClientInfo, params.Capabilities)
 
 	if options.ShowBugReports {
@@ -97,7 +104,7 @@ func (s *server) Initialize(ctx context.Context, params *protocol.ParamInitializ
 	}
 	s.pendingFolders = append(s.pendingFolders, folders...)
 
-	var codeActionProvider interface{} = true
+	var codeActionProvider any = true
 	if ca := params.Capabilities.TextDocument.CodeAction; len(ca.CodeActionLiteralSupport.CodeActionKind.ValueSet) > 0 {
 		// If the client has specified CodeActionLiteralSupport,
 		// send the code actions we support.
@@ -119,7 +126,7 @@ func (s *server) Initialize(ctx context.Context, params *protocol.ParamInitializ
 		}
 	}
 
-	var renameOpts interface{} = true
+	var renameOpts any = true
 	if r := params.Capabilities.TextDocument.Rename; r != nil && r.PrepareSupport {
 		renameOpts = protocol.RenameOptions{
 			PrepareProvider: r.PrepareSupport,
@@ -163,8 +170,8 @@ func (s *server) Initialize(ctx context.Context, params *protocol.ParamInitializ
 				Range: &protocol.Or_SemanticTokensOptions_range{Value: true},
 				Full:  &protocol.Or_SemanticTokensOptions_full{Value: true},
 				Legend: protocol.SemanticTokensLegend{
-					TokenTypes:     protocol.NonNilSlice(options.SemanticTypes),
-					TokenModifiers: protocol.NonNilSlice(options.SemanticMods),
+					TokenTypes:     moreslices.ConvertStrings[string](semtok.TokenTypes),
+					TokenModifiers: moreslices.ConvertStrings[string](semtok.TokenModifiers),
 				},
 			},
 			SignatureHelpProvider: &protocol.SignatureHelpOptions{
@@ -539,7 +546,8 @@ func (s *server) fetchFolderOptions(ctx context.Context, folder protocol.Documen
 
 	opts = opts.Clone()
 	for _, config := range configs {
-		s.handleOptionErrors(ctx, opts.Set(config))
+		res, errs := opts.Set(config)
+		s.handleOptionResult(ctx, res, errs)
 	}
 	return opts, nil
 }
@@ -553,7 +561,12 @@ func (s *server) eventuallyShowMessage(ctx context.Context, msg *protocol.ShowMe
 	s.notifications = append(s.notifications, msg)
 }
 
-func (s *server) handleOptionErrors(ctx context.Context, optionErrors []error) {
+func (s *server) handleOptionResult(ctx context.Context, applied []telemetry.CounterPath, optionErrors []error) {
+	for _, path := range applied {
+		path = append(settings.CounterPath{"gopls", "setting"}, path...)
+		counter.Inc(path.FullName())
+	}
+
 	var warnings, errs []string
 	for _, err := range optionErrors {
 		if err == nil {
