@@ -91,19 +91,19 @@ func TestOptions_Set(t *testing.T) {
 			},
 		},
 		{
-			name:      "hoverKind",
-			value:     "Structured",
-			wantError: true,
+			name:  "hoverKind",
+			value: "Structured",
+			// wantError: true, // TODO(rfindley): reinstate this error
 			check: func(o Options) bool {
-				return o.HoverKind == FullDocumentation
+				return o.HoverKind == Structured
 			},
 		},
 		{
-			name:      "ui.documentation.hoverKind",
-			value:     "Structured",
-			wantError: true,
+			name:  "ui.documentation.hoverKind",
+			value: "Structured",
+			// wantError: true, // TODO(rfindley): reinstate this error
 			check: func(o Options) bool {
-				return o.HoverKind == FullDocumentation
+				return o.HoverKind == Structured
 			},
 		},
 		{
@@ -180,6 +180,17 @@ func TestOptions_Set(t *testing.T) {
 				return len(o.DirectoryFilters) == 0
 			},
 		},
+		{
+			name: "annotations",
+			value: map[string]any{
+				"Nil":      false,
+				"noBounds": true,
+			},
+			wantError: true,
+			check: func(o Options) bool {
+				return !o.Annotations[Nil] && !o.Annotations[Bounds]
+			},
+		},
 		{
 			name:      "vulncheck",
 			value:     []any{"invalid"},
@@ -206,7 +217,7 @@ func TestOptions_Set(t *testing.T) {
 
 	for _, test := range tests {
 		var opts Options
-		err := opts.Set(map[string]any{test.name: test.value})
+		_, err := opts.Set(map[string]any{test.name: test.value})
 		if err != nil {
 			if !test.wantError {
 				t.Errorf("Options.set(%q, %v) failed: %v",
